/*
 This is part of the OTF library. Copyright by ZIH, TU Dresden 2005-2006.
 Authors: Andreas Knuepfer, Holger Brunst, Ronny Brendel, Thomas Kriebitzsch
*/

#include "config.h"

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <otf.h>

#include "handler.h"
#include "hash.h"


#define OTFMERGE_STRING "otfmerge"

#define HELPTEXT "" \
"                                                                  \n" \
" otfmerge - converter program of OTF library.                     \n" \
"                                                                  \n" \
" otfmerge [Options] <input file name>                             \n" \
"                                                                  \n" \
"   options:                                                       \n" \
"      -h, --help    show this help message                        \n" \
"      -n <n>        set number of streams for output              \n" \
"      -f <n>        set max number of filehandles available       \n" \
"      -o <name>     namestub of the output file (default 'out')   \n" \
"      -rb <size>    set buffersize of the reader                  \n" \
"      -wb <size>    set buffersize of the writer                  \n" \
"      -stats        cover statistics too                          \n" \
"      -snaps        cover snaphots too                            \n" \
"      -V            show OTF version                              \n" \
"      -z            write compressed output                       \n" \
"                                                                  \n" \
" Part of OTF library. Copyright by ZIH, TU Dresden, 2005-2006.    \n" \
"                                                                  \n"

int main ( int argc, char** argv ) {


	OTF_Reader* reader = NULL;
	OTF_HandlerArray* handlers;
	OTF_FileManager* manager;
	OTF_MasterControl* mc;
	OTF_MapEntry* mapentry;
	
	int read_stats= 0;
	int read_snaps= 0;
	int nstreams = 1;
	int nfiles = 1;
	int i;

	int readerbuffersize = 1024 * 1024;
	int writerbuffersize = 1024 * 1024;
	OTF_FileCompression compression= OTF_FILECOMPRESSION_UNCOMPRESSED;
	
	fcbT fcb;
	
	char* infile= NULL;
	char* outfile= NULL;
	

	if ( 1 >= argc ) {

			fprintf( stdout, HELPTEXT );
			return 0;
	}

	for ( i = 1; i < argc; i++ ) {


		if ( ( 0 == strcmp( "-i", argv[i] ) ) && ( i+1 < argc ) ) {

			infile= argv[i+1];
			++i;

		} else if ( ( 0 == strcmp( "-n", argv[i] ) ) && ( i+1 < argc ) ) {

			nstreams = atoi( argv[i+1] );
			++i;

		} else if ( ( 0 == strcmp( "-rb", argv[i] ) ) && ( i+1 < argc ) ) {

			readerbuffersize = atoi( argv[i+1] );
			++i;

		} else if ( ( 0 == strcmp( "-wb", argv[i] ) ) && ( i+1 < argc ) ) {

			writerbuffersize = atoi( argv[i+1] );
			++i;

		} else if ( ( 0 == strcmp( "-f", argv[i] ) ) && ( i+1 < argc ) ) {

			nfiles = atoi( argv[i+1] );
			++i;

		} else if ( ( 0 == strcmp( "-o", argv[i] ) ) && ( i+1 < argc ) ) {

			outfile= argv[i+1];
			++i;

		} else if ( 0 == strcmp( "-stats", argv[i] ) ) {

			read_stats= 1;

		} else if ( 0 == strcmp( "-snaps", argv[i] ) ) {

			read_snaps= 1;

		} else if ( 0 == strcmp( "-z", argv[i] ) ) {

			compression= OTF_FILECOMPRESSION_COMPRESSED;

		} else if ( 0 == strcmp( "-V", argv[i] ) ) {
		
			printf( "%u.%u.%u \"%s\"\n", OTF_VERSION_MAYOR, OTF_VERSION_MINOR,
				OTF_VERSION_SUB, OTF_VERSION_STRING);
			exit( 0 );

		} else if ( 0 == strcmp( "--help", argv[i] ) || 
				0 == strcmp( "-h", argv[i] ) ) {

			fprintf( stdout, HELPTEXT );
			return 0;

		} else {

			if ( '-' != argv[i][0] ) {

				infile= argv[i];

			} else{

				fprintf( stderr, "ERROR: Unknown option: '%s'\n", argv[i] );
				exit(1);
			}
		}
	}
	
	if ( nfiles < 1 ) {
	
		fprintf( stderr, "ERROR: less than 1 filehandle is not permitted\n" );
		exit(1);
	}
	if ( nstreams < 1 ) {
	
		fprintf( stderr, "ERROR: less than 1 stream is not permitted\n" );
		exit(1);
	}
	if ( NULL == infile ) {
	
		fprintf( stderr, "ERROR: no input file specified\n" );
		exit(1);
	}
	if ( NULL == outfile ) {

		/*
		fprintf( stderr, "ERROR: no output file has been specified\n" );
		exit(1);
		*/

		outfile= "out.otf";
	}

	handlers = OTF_HandlerArray_open();

	manager= OTF_FileManager_open( nfiles );
	assert( NULL != manager );
	
	reader = OTF_Reader_open( infile, manager );

	if ( NULL == reader ) {

		fprintf( stderr, "cannot open '%s'\n", infile );
		exit( 1 );
	}

	OTF_Reader_setBufferSizes( reader, readerbuffersize );
	
	fcb.writer = OTF_Writer_open( outfile, nstreams, manager );
	OTF_Writer_setBufferSizes( fcb.writer, writerbuffersize );
	OTF_Writer_setCompression( fcb.writer, compression );
	
	mc = OTF_Reader_getMasterControl( reader );

	/* set your own handler functions */

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefinitionComment,
		OTF_DEFINITIONCOMMENT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFINITIONCOMMENT_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefTimerResolution,
		OTF_DEFTIMERRESOLUTION_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFTIMERRESOLUTION_RECORD);

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefProcess,
		OTF_DEFPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFPROCESS_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefProcessGroup,
		OTF_DEFPROCESSGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFPROCESSGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefFunction,
		OTF_DEFFUNCTION_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFFUNCTION_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefFunctionGroup,
		OTF_DEFFUNCTIONGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFFUNCTIONGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefCounter,
		OTF_DEFCOUNTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFCOUNTER_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefCounterGroup,
		OTF_DEFCOUNTERGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFCOUNTERGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefScl,
		OTF_DEFSCL_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFSCL_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefSclFile,
		OTF_DEFSCLFILE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFSCLFILE_RECORD );

/*	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefversion,
		OTF_DEFVERSION_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFVERSION_RECORD );
*/
	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleDefCreator,
		OTF_DEFCREATOR_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_DEFCREATOR_RECORD );



	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleEventComment,
		OTF_EVENTCOMMENT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_EVENTCOMMENT_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleCounter,
		OTF_COUNTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_COUNTER_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleEnter,
		OTF_ENTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_ENTER_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleCollectiveOperation,
		OTF_COLLOP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_COLLOP_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleRecvMsg,
		OTF_RECEIVE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_RECEIVE_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleSendMsg,
		OTF_SEND_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_SEND_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleLeave,
		OTF_LEAVE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_LEAVE_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleBeginProcess,
		OTF_BEGINPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, &fcb, 
		OTF_BEGINPROCESS_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleEndProcess,
		OTF_ENDPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, &fcb, 
		OTF_ENDPROCESS_RECORD );
			

	/* snapshot records */

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleEnterSnapshot,
		OTF_ENTERSNAPSHOT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, &fcb, 
		OTF_ENTERSNAPSHOT_RECORD );
	
	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleSendSnapshot,
		OTF_SENDSNAPSHOT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, &fcb, 
		OTF_SENDSNAPSHOT_RECORD );	


	/* summary records */

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleFunctionSummary,
		OTF_FUNCTIONSUMMARY_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_FUNCTIONSUMMARY_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleFunctionGroupSummary,
		OTF_FUNCTIONGROUPSUMMARY_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_FUNCTIONGROUPSUMMARY_RECORD );

	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleMessageSummary,
		OTF_MESSAGESUMMARY_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_MESSAGESUMMARY_RECORD );

	/* misc records */
	
	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) OTF_handleUnknown,
		OTF_UNKNOWN_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, 
        &fcb, OTF_UNKNOWN_RECORD );

	/* ask the mastercontrol for the number of streams and create the
	 * streaminfos array
	 */
	fcb.nstreaminfos = 0;
	
	while ( 0 !=  OTF_MasterControl_getEntryByIndex( mc, fcb.nstreaminfos )) {
	
		fcb.nstreaminfos++;
	}
	
	fcb.hash = hash_new ();
	
	/* global stream yourself, because he isnt in the mapping */
	hash_add( fcb.hash, 0 );
	
	/* add all streams to the hash */
	for( i = 0; i < fcb.nstreaminfos; i++ ) {
	
		mapentry = OTF_MasterControl_getEntryByIndex( mc, i );
		
		hash_add( fcb.hash, mapentry->argument );
	}
	
	/* read definitions */
	OTF_Reader_readDefinitions( reader, handlers );
	
	/* --------------------------------------------------- */
	/* output all functions and functiongroups -- start -- */
	/*	si = hash_search( fcb.hash, 0 );
	
	/ * stream zero * /
	if ( si->nfunctiongroups ) {
	
		fprintf( stderr, "stream %u---\n", 0 );
		
		for( a = 0; a < si->nfunctiongroups; a++ ) {
		
			fprintf( stderr, "  DFG token: %u name: \"%s\"\n",
				si->functiongroups[a].deftoken,
				si->functiongroups[a].name );
		}
	}
	
	if ( si->nfunctions > 0 ) {
		fprintf( stderr, "stream %u---\n", 0 );
		
		for( a = 0; a < si->nfunctions; a++ ) {
		
			fprintf( stderr, "  DF token: %u group: %u name: \"%s\" scl: %u\n",
				si->functions[a].deftoken,
				si->functions[a].group,
				si->functions[a].name,
				si->functions[a].scltoken );
		}	
	}
		
	/ * all other streams * /
	for( i = 0; i < fcb.nstreaminfos; i++ ) {
	
		mapentry = OTF_MasterControl_getEntryByIndex( mc, i );
		
		si = hash_search( fcb.hash, mapentry->argument );
		
		if ( si->nfunctiongroups ) {
		
			fprintf( stderr, "stream %u---\n", mapentry->argument );
			
			for( a = 0; a < si->nfunctiongroups; a++ ) {
			
				fprintf( stderr, "  DFG token: %u name: \"%s\"\n",
					si->functiongroups[a].deftoken,
					si->functiongroups[a].name );
			}
		}
		
		if ( si->nfunctions > 0 ) {
			fprintf( stderr, "stream %u---\n", mapentry->argument );
			
			for( a = 0; a < si->nfunctions; a++ ) {
			
				fprintf( stderr, "  DF token: %u group: %u name: \"%s\" scl: %u\n",
					si->functions[a].deftoken,
					si->functions[a].group,
					si->functions[a].name,
					si->functions[a].scltoken );
			}
		}
	}
	*/	
	/* output all functions and functiongroups -- end -- */
	/* ------------------------------------------------- */
	
	

	/* read events */
	OTF_Reader_readEvents( reader, handlers );
	
	if ( 1 == read_stats ) {
	
		OTF_Reader_readStatistics( reader, handlers );
	
	}
	
	if ( 1 == read_snaps ) {
	
		OTF_Reader_readSnapshots( reader, handlers );
		
	}

	OTF_Writer_close( fcb.writer );
	OTF_Reader_close( reader );
	OTF_FileManager_close( manager );
	OTF_HandlerArray_close( handlers );

	return 0;
}
